// license:BSD-3-Clause
// copyright-holders:Fabrice Bellet, Olivier Galibert

// Management of Thomson Basic floppy images

#include "fs_thom_basic.h"
#include "fsblk.h"
#include "thom_dsk.h"

#include "multibyte.h"
#include "strformat.h"

#include <stdexcept>

using namespace fs;

namespace fs { const thomson_basic_image THOMSON_BASIC; }

// Floppy only, format is 40/80 tracks, 1 head, 16 sectors.
// Filesystem has no subdirectories.


// Track 20 sector 2 has the inode table (8 sectors per block)
//    offset 0: 0
//    offset 1+(track*2) if sector <= 8
//    offset 2+(track*2) if sector > 8
//
//  127 inodes in single density (sector size 128)
//  160 inodes in double density (sector size 256, 80 tracks)
//
// Inode:
//   ff: block unallocated
//   fe: block reserved
//   00-bf: allocated blocks, next inode
//   c1-c8: last block, number of sectors used

// Track 20 sector 3-16: directory sectors
// Directory sector format (32 bytes per file):
//  offset 00-07: filename, left justified, space padded
//      00:    unallocated
//      20-7f: allocated entry (filename first character)
//      ff:    last entry
//  offset 08-0a: suffix, left justified, space padded
//  offset 0b:    filetype
//	0: Basic program ASCII or binary
//      1: Basic Data ASCII
//      2: machine code binary
//      3: assembler ASCII
//  offset 0c:    flag ASCII (00) or binary (ff)
//  offset 0d:    first inode
//  offset 0e-0f: number of bytes used in the last inode
//  offset 10-17: comments
//  offset 18-1f: reserved
//

namespace {

class thomson_basic_impl : public filesystem_t {
public:
	thomson_basic_impl(fsblk_t &blockdev);
	virtual ~thomson_basic_impl() = default;

	virtual meta_data volume_metadata() override;
	virtual std::pair<err_t, meta_data> metadata(const std::vector<std::string> &path) override;
	virtual err_t volume_metadata_change(const meta_data &info) override;

	virtual std::pair<err_t, std::vector<dir_entry>> directory_contents(const std::vector<std::string> &path) override;
	virtual err_t rename(const std::vector<std::string> &opath, const std::vector<std::string> &npath) override;
	virtual err_t remove(const std::vector<std::string> &path) override;

	virtual err_t file_create(const std::vector<std::string> &path, const meta_data &meta) override;

	virtual std::pair<err_t, std::vector<u8>> file_read(const std::vector<std::string> &path) override;
	virtual err_t file_write(const std::vector<std::string> &path, const std::vector<u8> &data) override;

	virtual err_t format(const meta_data &meta) override;

	static bool validate_filename(std::string name);

private:
	int m_block_size = 0;
	u8 m_sect_size = 0;
	u8 m_track_count = 0;

	std::vector<u8> allocate_blocks(u8 nblocks, u8 nsects, u8 nbytes);
	void free_blocks(u8 first_block);
	u8 free_block_count();

	static std::string file_name_read(const u8 *p);
	static std::string file_name_prepare(std::string name);
	meta_data file_metadata(const u8 *entry);
	std::pair<fsblk_t::block_t, u32> file_find(std::string name);
};
}

const char *thomson_basic_image::name() const
{
	return "thomson_basic";
}

const char *thomson_basic_image::description() const
{
	return "Thomson Basic";
}

void thomson_basic_image::enumerate_f(floppy_enumerator &fe) const
{
	fe.add(FLOPPY_THOMSON_FORMAT, floppy_image::FF_35, floppy_image::SSDD, 327680, "tb35ssdd", "Thomson Basic 3.5\" single-sided double-density");
	fe.add(FLOPPY_THOMSON_FORMAT, floppy_image::FF_525, floppy_image::SSDD, 163840, "tb525ssdd", "Thomson Basic 5.25\" single-sided double-density");
	fe.add(FLOPPY_THOMSON_FORMAT, floppy_image::FF_525, floppy_image::SSSD, 81920, "tb525sssd", "Thomson Basic 5.25\" single-sided single-density");
}

std::unique_ptr<filesystem_t> thomson_basic_image::mount(fsblk_t &blockdev) const
{
	return std::make_unique<thomson_basic_impl>(blockdev);
}

bool thomson_basic_image::can_format() const
{
	return true;
}

bool thomson_basic_image::can_read() const
{
	return true;
}

bool thomson_basic_image::can_write() const
{
	return true;
}

bool thomson_basic_image::has_rsrc() const
{
	return false;
}

std::vector<meta_description> thomson_basic_image::volume_meta_description() const
{
	std::vector<meta_description> res;
	res.emplace_back(meta_description(meta_name::name, "NoName", false, [](const meta_value &m) { return m.as_string().size() <= 8; }, "Volume name, up to 8 characters"));

	return res;
}

std::vector<meta_description> thomson_basic_image::file_meta_description() const
{
	std::vector<meta_description> res;
	res.emplace_back(meta_description(meta_name::name, "", false, [](const meta_value &m) { return thomson_basic_impl::validate_filename(m.as_string()); }, "File name, 8.3"));
	res.emplace_back(meta_description(meta_name::length, 0, true, nullptr, "Size of the file in bytes"));
	res.emplace_back(meta_description(meta_name::size_in_blocks, 0, true, nullptr, "Number of blocks used by the file"));
	res.emplace_back(meta_description(meta_name::file_type, "", false, nullptr, "File Type"));
	res.emplace_back(meta_description(meta_name::ascii_flag, "", false, nullptr, "File Format"));
	res.emplace_back(meta_description(meta_name::comments, "", false, [](const meta_value &m) { return m.as_string().size() <= 8; }, "Comments, up to 8 characters"));
	return res;
}

thomson_basic_impl::thomson_basic_impl(fsblk_t &blockdev) : filesystem_t(blockdev, 256)
{
	switch(blockdev.block_count()) {
		case 1280:
			m_track_count = 80;
			m_sect_size = 255;
			m_block_size = 256;
			break;
		case 640:
			m_track_count = 40;
			m_sect_size = 255;
			m_block_size = 256;
			break;
		case 320:
			m_track_count = 40;
			m_sect_size = 128;
			m_block_size = 128;
			blockdev.set_block_size(128);
	}
}

bool thomson_basic_impl::validate_filename(std::string name)
{
	auto pos = name.find('.');
	if(pos != std::string::npos)
		return pos <= 8 && pos > 0 && name.size()-pos-1 <= 3;
	else
		return name.size() > 0 && name.size() <= 8;
}

err_t thomson_basic_impl::format(const meta_data &meta)
{
	std::string volume_name = meta.get_string(meta_name::name, "NoName");

	m_blockdev.fill(0xe5);

	auto bvol = m_blockdev.get(20*16);
	volume_name.resize(8, ' ');
	bvol.wstr(0, volume_name);

	auto fmap = m_blockdev.get(20*16 + 1);
	fmap.fill(0xff);
	fmap.w8(0, 0x00);
	fmap.w8(20*2 + 1, 0xfe);
	fmap.w8(20*2 + 2, 0xfe);

	for(int i=2; i < 16; i++) {
		auto bdir = m_blockdev.get(20*16 + i);
		bdir.fill(0xff);
	}

	return ERR_OK;
}

meta_data thomson_basic_impl::volume_metadata()
{
	meta_data res;
	auto bvol = m_blockdev.get(20*16);
	int len = 8;
	while(len > 0 && bvol.rodata()[len - 1] == ' ')
		len--;

	res.set(meta_name::name, bvol.rstr(0, len));
	return res;
}

err_t thomson_basic_impl::volume_metadata_change(const meta_data &meta)
{
	if(meta.has(meta_name::name)) {
		std::string volume_name = meta.get_string(meta_name::name);
		volume_name.resize(8, ' ');
		m_blockdev.get(20*16).wstr(0, volume_name);
	}
	return ERR_OK;
}

std::string thomson_basic_impl::file_name_prepare(std::string fname)
{
	std::string nname;
	size_t i;
	for(i = 0; i != 8 && i != fname.size() && fname[i] != '.'; i++)
		nname += fname[i];
	if(nname.size() != 8)
		nname.insert(nname.end(), 8 - nname.size(), ' ');
	while(i != fname.size() && fname[i] != '.')
		i++;
	if(i != fname.size())
		i++;
	while(i != fname.size() && nname.size() != 11)
		nname += fname[i++];
	if(nname.size() != 11)
		nname.insert(nname.end(), 11 - nname.size(), ' ');
	return nname;
}

std::string thomson_basic_impl::file_name_read(const u8 *p)
{
	int main_len;
	for(main_len = 8; main_len > 0; main_len--)
		if(p[main_len - 1] != ' ')
			break;
	int ext_len;
	for(ext_len = 3; ext_len > 0; ext_len--)
		if(p[ext_len + 8 - 1] != ' ')
			break;
	std::string name;
	for(int i=0; i != main_len; i++)
		name += char(p[i]);
	name += '.';
	for(int i=0; i != ext_len; i++)
		name += char(p[8 + i]);
	return name;
}

meta_data thomson_basic_impl::file_metadata(const u8 *entry)
{
	meta_data res;
	static char type[5] = "BDMA";

	res.set(meta_name::name, file_name_read(entry));

	if (entry[0xb] <= 3)
		res.set(meta_name::file_type, util::string_format("%c", type[entry[0xb]]));
	if (entry[0xc] == 0 || entry[0xc] == 0xff)
		res.set(meta_name::ascii_flag, util::string_format(entry[0xc] == 0xff ? "A" : "B"));

	u8 b = entry[0xd];
	u32 byte_count = 0;
	u8 block_count = 1;
	auto fmap = m_blockdev.get(20*16 + 1);
	u8 next = fmap.r8(b + 1);
	while (next < 0xc0) {
		block_count++;
		byte_count += 8 * m_sect_size;
		b = next;
		next = fmap.r8(next + 1);
	}
	if (get_u16be(entry + 0xe) > 0) {
		byte_count += (next - 0xc1) * m_sect_size + get_u16be(entry + 0xe);
	} else
		byte_count += (next - 0xc0) * m_sect_size;

	res.set(meta_name::size_in_blocks, block_count);
	res.set(meta_name::length, byte_count);

	return res;
}

std::pair<fsblk_t::block_t, u32> thomson_basic_impl::file_find(std::string name)
{
	name = file_name_prepare(name);
	for(u8 i = 2; i < 16; i++) {
		auto bdir = m_blockdev.get(20*16 + i);
		for(int j = 0; j < m_block_size; j += 32) {
			if (bdir.r8(j) == 0x00)
				continue;
			if (bdir.r8(j) == 0xff)
				break;
			if(memcmp(bdir.rodata() + j, name.data(), 11) == 0)
				return std::make_pair(bdir, j);
		}
	}
	return std::make_pair(nullptr, 0xff);
}

std::pair<err_t, meta_data> thomson_basic_impl::metadata(const std::vector<std::string> &path)
{
	if(path.size() != 1)
		return std::make_pair(ERR_NOT_FOUND, meta_data());

	auto [bdir, off] = file_find(path[0]);
	if(off == 0xff)
		return std::make_pair(ERR_NOT_FOUND, meta_data());

	return std::make_pair(ERR_OK, file_metadata(bdir.rodata() + off));
}

std::pair<err_t, std::vector<dir_entry>> thomson_basic_impl::directory_contents(const std::vector<std::string> &path)
{
	std::pair<err_t, std::vector<dir_entry>> res;

	if(path.size() != 0) {
		res.first = ERR_NOT_FOUND;
		return res;
	}

	res.first = ERR_OK;

	for(u8 i = 2; i < 16; i++) {
		auto bdir = m_blockdev.get(20*16 + i);
		for(int j = 0; j < m_block_size; j += 32) {
			if (bdir.r8(j) == 0x00)
				continue;
			if (bdir.r8(j) == 0xff)
				break;
			meta_data meta = file_metadata(bdir.rodata() + j);
			res.second.emplace_back(dir_entry(dir_entry_type::file, meta));
		}
	}
	return res;
}

err_t thomson_basic_impl::rename(const std::vector<std::string> &opath, const std::vector<std::string> &npath)
{
	if(opath.size() != 1 || npath.size() != 1)
		return ERR_NOT_FOUND;

	auto [bdir, off] = file_find(opath[0]);
	if(off == 0xff)
		return ERR_NOT_FOUND;

	wstr(bdir.data() + off, file_name_prepare(npath[0]));

	return ERR_OK;
}

err_t thomson_basic_impl::remove(const std::vector<std::string> &path)
{
	if(path.size() != 1)
		return ERR_NOT_FOUND;

	auto [bdir, off] = file_find(path[0]);
	if(off == 0xff)
		return ERR_NOT_FOUND;

	u8 *entry = bdir.data() + off;
	free_blocks(entry[0xd]);
	bdir.w8 (off, 0xff);
	return ERR_OK;
}

err_t thomson_basic_impl::file_create(const std::vector<std::string> &path, const meta_data &meta)
{
	if(path.size() != 0)
		return ERR_NOT_FOUND;

	auto bdir = m_blockdev.get(20*16 + 2);
	int off = 0;
	for(u8 i = 2; i < 16; i++) {
		bdir = m_blockdev.get(20*16 + i);
		for(off = 0; off < m_block_size; off += 32) {
			if (bdir.r8(off) == 0x00)
				goto found;
			if (bdir.r8(off) == 0xff)
				goto found;
		}
	}
	return ERR_NO_SPACE;
found:
	auto blocks = allocate_blocks(0, 0, 1);
	if(blocks.empty()) {
		remove(path);
		return ERR_NO_SPACE;
	}

	bdir.wstr(off, file_name_prepare(meta.get_string(meta_name::name, "")));
	bdir.w8  (off+0xb, 0);    // type: Basic program
	bdir.w8  (off+0xc, 0xff); // format: ASCII
	bdir.w8  (off+0xd, blocks[0]);
	bdir.w16b(off+0xe, 1);    // dummy size
	for(u8 i = 0x10; i < 0x20; i++)
		bdir.w8 (off+i, 0);

	return ERR_OK;
}

std::pair<err_t, std::vector<u8>> thomson_basic_impl::file_read(const std::vector<std::string> &path)
{
	std::vector<u8> data;

	if(path.size() != 1)
		return std::make_pair(ERR_NOT_FOUND, data);

	auto [bdir, off] = file_find(path[0]);
	if(off == 0xff)
		return std::make_pair(ERR_NOT_FOUND, data);

	const u8 *entry = bdir.rodata() + off;
	u8 b = entry[0xd];
	auto fmap = m_blockdev.get(20*16 + 1);
	u8 next = fmap.r8(b + 1);
	u32 pos = 0;
	while(next < 0xc0) {
		for(u8 j = 0; j < 8; j++) {
			data.resize(pos + m_sect_size);
			auto dblk = m_blockdev.get(8 * b + j);
			memcpy(data.data() + pos, dblk.rodata(), m_sect_size);
			pos += m_sect_size;
		}
		b = next;
		next = fmap.r8(b + 1);
	}
	u8 sect_count = next - 0xc0;
	if (get_u16be(entry + 0xe) > 0)
		sect_count--;
	for(u8 j = 0; j < sect_count; j++) {
		data.resize(pos + m_sect_size);
		auto dblk = m_blockdev.get(8 * b + j);
		memcpy(data.data() + pos, dblk.rodata(), m_sect_size);
		pos += m_sect_size;
	}
	if(get_u16be(entry + 0xe) > 0) {
		data.resize(pos + get_u16be(entry + 0xe));
		auto dblk = m_blockdev.get(8 * b + next - 0xc1);
		memcpy(data.data() + pos, dblk.rodata(), get_u16be(entry + 0xe));
	}
	return std::make_pair(ERR_OK, data);
}

err_t thomson_basic_impl::file_write(const std::vector<std::string> &path, const std::vector<u8> &data)
{
	int nblocks = data.size() / (8 * m_sect_size);
	int nsects = (data.size() % (8 * m_sect_size)) / m_sect_size;
	int nbytes = data.size() - nblocks * 8 * m_sect_size - nsects * m_sect_size;

	if(path.size() != 1)
		return ERR_NOT_FOUND;

	auto [bdir, off] = file_find(path[0]);
	if(off == 0xff)
		return ERR_NOT_FOUND;

	u8 *entry = bdir.data() + off;
	free_blocks(entry[0xd]);
	auto blocks = allocate_blocks (nblocks, nsects, nbytes);
	if (blocks.empty()) {
		remove(path);
		return ERR_NO_SPACE;
	}
	bdir.w8  (off+0xd, blocks[0]);
	bdir.w16b(off+0xe, nbytes);

	u32 pos = 0;
	for(u8 i = 0; i < nblocks; i++) {
		for (u8 j = 0; j < 8; j++) {
			auto dblk = m_blockdev.get(8 * blocks[i] + j);
			dblk.copy(0, data.data() + pos, m_sect_size);
			pos += m_sect_size;
		}
	}
	for(u8 j = 0; j < nsects; j++) {
		auto dblk = m_blockdev.get(8 * blocks.back() + j);
		dblk.copy(0, data.data() + pos, m_sect_size);
		pos += m_sect_size;
	}
	if(nbytes > 0) {
		auto dblk = m_blockdev.get(8 * blocks.back() + nsects);
		dblk.copy(0, data.data() + pos, data.size() - pos);
	}
	return ERR_OK;
}

std::vector<u8> thomson_basic_impl::allocate_blocks(u8 nblocks, u8 nsects, u8 nbytes)
{
	if (nbytes > 0)
		nsects++;
	if (nsects > 0)
		nblocks++;

	std::vector<u8> blocks;

	if(free_block_count() < nblocks)
		return blocks;

	if(nblocks == 0)
		return blocks;

	auto fmap = m_blockdev.get(20*16 + 1);
	u8 nf = 0;
	u8 block_count = (m_track_count * 2 < m_block_size - 1 ? m_track_count * 2 : m_block_size - 1);
	for(u8 i = 0; i < block_count; i++) {
		if (fmap.r8(i + 1) == 0xff) {
			nf++;
			if(blocks.size() > 0)
				fmap.w8(blocks.back() + 1, i);
			blocks.push_back(i);
		}
		if(nf == nblocks)
			break;
	}
	if (nsects == 0)
		nsects = 8;
	fmap.w8(blocks.back() + 1, 0xc0 + nsects);
	return blocks;
}

void thomson_basic_impl::free_blocks(u8 first_block)
{
	auto fmap = m_blockdev.get(20*16 + 1);
	u8 b = first_block;
	while (true) {
		u8 next = fmap.r8(b + 1);
		if (next == 0xff || next == 0xfe)
			return;
		fmap.w8(b + 1, 0xff);
		if (next >= 0xc1 && next <= 0xc8)
			return;
		b = next;
	}
}

u8 thomson_basic_impl::free_block_count()
{
	auto fmap = m_blockdev.get(20*16 + 1);
	u32 nf = 0;
	u8 block_count = (m_track_count * 2 < m_block_size - 1 ? m_track_count * 2 : m_block_size - 1);
	for(u8 i = 1; i <= block_count; i++) {
		if (fmap.r8(i) == 0xff)
			nf++;
	}
	return nf;
}
