// license:BSD-3-Clause
// copyright-holders: Fabrice Bellet

#include <assert.h>
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include "corestr.h"
#include "ioprocs.h"
#include "formats/cassimg.h"

#define VERBOSE		0 /* 0, 1, 2 or 3 */

#define LOG(x)		do { if(VERBOSE > 0) printf x; } while(0)
#define VLOG(x)		do { if(VERBOSE > 1) printf x; } while(0)
#define VVLOG(x)	do { if(VERBOSE > 2) printf x; } while(0)

// see src/mame/thomson/thomson_m.cpp for the description of the format
#define TO7_BIT_LENGTH	0.001114
#define MO5_BIT_LENGTH	0.000833
#define MO5_HBIT_LENGTH	(MO5_BIT_LENGTH / 2.)

static void usage(void)
{
	fprintf(stderr, "wav2k7 - Thomson wave to cassette conversion tool for use with MAME\n\n");
	fprintf(stderr, "Usage:\n");
	fprintf(stderr, "	wav2k7 convert <format> <inputfile.wav> <outputfile>\n\n");
	fprintf(stderr, "Supported formats:\n\n");
	fprintf(stderr, "       mo5 - Thomson MO-series\n");
	fprintf(stderr, "                k5,k7\n");
	fprintf(stderr, "       to7 - Thomson TO-series\n");
	fprintf(stderr, "                k7\n\n");
	fprintf(stderr, "Example usage:\n");
	fprintf(stderr, "        wav2k7 convert to7 game.wav game.k7\n\n");
}

#define same_sign(x,y)  (((x) >= 0 && (y) >= 0) || ((x) < 0 && (y) < 0))

static double mo5_next_sign_change(cassette_image::ptr &cassette, double pos)
{
	cassette_image::Info info = cassette->get_info();
	double len = (double)info.sample_count / info.sample_frequency;
	int8_t data, next;

	if(pos < len) {
		cassette->get_samples(0, pos, MO5_BIT_LENGTH, 1, 1, &data, 0);
		pos += MO5_HBIT_LENGTH / 20;
	}
	while(pos < len) {
		cassette->get_samples(0, pos, MO5_BIT_LENGTH, 1, 1, &next, 0);
		if(!same_sign(data, next))
			return pos;
		pos += MO5_HBIT_LENGTH / 20;
		data = next;
	}
	return pos;
}

static uint8_t mo5_next_bit(cassette_image::ptr &cassette, double &pos, bool &error)
{
	cassette_image::Info info = cassette->get_info();
	double len = (double)info.sample_count / info.sample_frequency;
	double period;
	double next;

#define IS_HBIT(v) ((v) > 0.8 * MO5_HBIT_LENGTH)
#define IS_BIT(v) ((v) > 0.8 * MO5_BIT_LENGTH)

	next = mo5_next_sign_change(cassette, pos);
	period = next - pos;
	if(IS_BIT(period)) {
		error = (period > 1.2 * MO5_BIT_LENGTH);
		pos = (error ? pos + MO5_BIT_LENGTH : next);
		return 0;
	}
	if(next < len && IS_HBIT(period)) {
		next = mo5_next_sign_change(cassette, next);
		period = next - pos;
		if(IS_BIT(period)) {
			error = (period > 1.2 * MO5_BIT_LENGTH);
			pos = (error ? pos + MO5_BIT_LENGTH : next);
			return 1;
		}
	}
	pos += MO5_BIT_LENGTH;
	error = true;
	return 0;
}

static uint8_t mo5_next_byte(cassette_image::ptr &cassette, double &pos, uint8_t &nbits)
{
	cassette_image::Info info = cassette->get_info();
	double len = (double)info.sample_count / info.sample_frequency;
	double start_pos = pos;
	static bool synced = false;
	uint8_t res = 0;
	bool error;

retry:
	if(!synced) {
		nbits = 0;
		pos = mo5_next_sign_change(cassette, pos);
		// skip zero filler
		while(pos < len && (mo5_next_bit(cassette, pos, error) == 0)) {
			nbits++;
			if(error)
				goto retry;
		}
		nbits++;
		synced = true;
		if(nbits > 7) {
			// the first bit 1 after enough bits 0 make
			// a valid first byte 0x01
			nbits = 8;
			VVLOG(("mo5_next_byte: %f - %f byte=01 (%d)\n", start_pos, pos, nbits));
			return 0x01;
		}
	}
	for(nbits = 0; nbits < 8 && pos < len; nbits++) {
		res = (res << 1) | mo5_next_bit(cassette, pos, error);
		if(error && nbits < 7) {
			// discard a byte crossing an empty area
			synced = false;
			goto retry;
		}
	}
	VVLOG(("mo5_next_byte: %f - %f byte=%02x (%d)\n", start_pos, pos, res, nbits));
	return res;
}

static int do_convert_mo5(cassette_image::ptr &cassette, FILE *f)
{
	cassette_image::Info info = cassette->get_info();
	double len = (double)info.sample_count / info.sample_frequency;
	double pos = 0;
	uint8_t data[260];
	bool valid = true;

	while(pos < len) {
		uint8_t nbits;
		data[0] = mo5_next_byte(cassette, pos, nbits);
		if(nbits < 8)
			break;

		int one_filler = 0;
		if(valid && data[0] == 0xdc) {
			LOG(( "do_convert_mo5: DC signature found at pos=%f\n",pos));
			data[0] = mo5_next_byte(cassette, pos, nbits);
			one_filler++;
		}

		while(nbits == 8 && data[0] == 0x01) {
			data[0] = mo5_next_byte(cassette, pos, nbits);
			one_filler++;
		}

		int index = 1;
		while(nbits == 8 && index < 4)
			if (data[0] != 0x3c)
				break;
			else if(index > 1 && data[1] != 0x5a)
				break;
			else if(index > 2 && data[2] != 0x00 && data[2] != 0x01 && data[2] != 0xff)
				break;
			else
				data[index++] = mo5_next_byte(cassette, pos, nbits);

		valid = (index == 4);

		int size = 0;
		if(valid && nbits == 8) {
			uint8_t csum = 0;
			size = (uint8_t)(data[3] - 1);
			while(nbits == 8 && index < size + 4) {
				data[index] = mo5_next_byte(cassette, pos, nbits);
				csum += data[index++];
			}

			if(index < size + 4)
				valid = false;
			else if(csum != 0)
				LOG(("do_convert_mo5: invalid checksum %02x (should be 0)\n", csum));
		}

		for(int i = 0; i < one_filler ; i++)
			fputc(1, f);
		if(one_filler > 0) {
			LOG(("do_convert_mo5: 0x01 filler size=%d\n", one_filler));
		}
		if(valid) {
			LOG(("do_convert_mo5: block type=$%02x size=%d\n", data[2], size));
			VLOG(("do_convert_mo5: data="));
		} else
			VLOG(("do_convert_mo5: special data="));
		for(int i = 0; i < index ; i++) {
			fputc(data[i], f);
			VLOG(("%02x ", data[i]));
		}
		VLOG(("\n"));

		if(valid && data[2] == 0) {
			char name[] = "01234567.ABC";
                        uint8_t t = data[15];
                        uint8_t u = data[16];
                        memcpy(name, data + 4, 8);
                        memcpy(name + 9, data + 12, 3);
                        for (index = 0; name[index]; index++)
                                if(name[index] < ' ' || name[index] >= 127)
                                        name[index] = '?';
                        LOG(( "do_convert_mo5: file \"%s\" type=%s,%s at %imn %is\n",
					name,
					(t==0) ? "bas" : (t==1) ? "dat" : (t==2) ? "bin" : "???",
					(u == 0) ? "a" : (u == 0xff) ? "b" : "?",
					(int)pos / 60, (int)pos % 60));
                }
	}
	return 0;
}

#define MAX(x,y) ((x) > (y) ? (x) : (y))

static uint8_t to7_next_bit(cassette_image::ptr &cassette, double &pos, bool &error)
{
	cassette_image::Info info = cassette->get_info();
	double len = (double)info.sample_count / info.sample_frequency;

	static int bitpos = 0;
	static bool synced = false;
	while(pos < len) {
		// demodulate wave signal on-the-fly
		// we simply count sign changes...
		int bit, chg = 0, nochg = 0, maxnochg = 0;
		int8_t data[40];
		cassette->get_samples(0, pos, TO7_BIT_LENGTH * 15. / 14., 40, 1, data, 0);
		
		for(int i = 1; i < 40; i++) {
			if(same_sign(data[i - 1], data[i]))
				nochg++;
			else {
				maxnochg = MAX(maxnochg, nochg);
				nochg = 0;
				chg++;
			}
		}
		maxnochg = MAX(maxnochg, nochg);
		bit = (chg >= 13 ? 1 : 0);

		VVLOG(("to7_next_bit: pos=%f samppos=%d bit%d=%d (%d-%d) synced=%d\n", pos, (int)(pos / TO7_BIT_LENGTH), bitpos, bit, chg, maxnochg, synced));

		if(maxnochg > 5) {
			pos += TO7_BIT_LENGTH;
			goto error;
		}

		if(!synced) {
			// when not in sync, look for the next bit 0
			if(bit)
				pos += TO7_BIT_LENGTH / 20;
			else {
				// we found the 0 start bit
				// change state to synchronized and
				// realign position index
				synced = 1;
				bitpos = 0;
				pos += (TO7_BIT_LENGTH / 2);
			}
		} else {
			// when in sync, progress by step of size TO7_BIT_LENGTH
			if(bitpos == 0) {
				// this bit is the 0 start bit, skip it
				pos += TO7_BIT_LENGTH;
				bitpos++;
				continue;
			} else if(bitpos >=1 && bitpos <= 8) {
				// this bit is part of the data
				pos += TO7_BIT_LENGTH;
				bitpos++;
				if(bitpos == 9)
					// we skip the next two 1 stop bits and
					// look for the next 0 start bit
					synced = false;
				error = false;
				return bit;
			}
		}
	}
error:
	error = true;
	synced = false;
	return 0;
}

static uint8_t to7_next_byte(cassette_image::ptr &cassette, double &pos, uint8_t &nbits)
{
	cassette_image::Info info = cassette->get_info();
	double len = (double)info.sample_count / info.sample_frequency;
	uint8_t res = 0;
	double start_pos = pos;

retry:
	for(nbits = 0; nbits < 8 && pos < len; nbits++) {
		bool error = false;
		res = (res >> 1) | (to7_next_bit(cassette, pos, error) ? 0x80 : 0);
		if(error && nbits < 7)
			// discard a byte crossing an empty area
			goto retry;
	}
	VVLOG(("to7_next_byte: %f - %f byte=%02x (%d)\n", start_pos, pos, res, nbits));
	return res;
}

static int do_convert_to7(cassette_image::ptr &cassette, FILE *f)
{
	cassette_image::Info info = cassette->get_info();
	double len = (double)info.sample_count / info.sample_frequency;
	double pos = 0;
	uint8_t data[260];
	int size = 0;
	bool valid = true;
	int ff_filler = 0;

	while(pos < len) {
		uint8_t nbits;
		data[0] = to7_next_byte(cassette, pos, nbits);
		if(nbits < 8)
			break;

		ff_filler = 0;
		if(valid && data[0] == 0xdc) {
			LOG(( "do_convert_to7: DC signature found at pos=%f\n",pos));
			data[0] = to7_next_byte(cassette, pos, nbits);
			ff_filler++;
			if(pos >= len)
				break;
		}

		while(nbits == 8 && data[0] == 0xff) {
			data[0] = to7_next_byte(cassette, pos, nbits);
			ff_filler++;
		}

		int index = 1;
		while(nbits == 8 && index < 4)
			if (data[0] != 0x01)
				break;
			else if(index > 1 && data[1] != 0x3c)
				break;
			else if(index > 2 && data[2] != 0x00 && data[2] != 0x01 && data[2] != 0xff)
				break;
			else
				data[index++] = to7_next_byte(cassette, pos, nbits);

		valid = (index == 4);

		if(valid && nbits == 8) {
			uint8_t csum = data[2] + data[3];
			size = (uint8_t)(data[3]);
			while(nbits == 8 && index < size + 4) {
				data[index] = to7_next_byte(cassette, pos, nbits);
				csum += data[index++];
			}
			if(nbits == 8 && index < size + 5) {
				data[index] = to7_next_byte(cassette, pos, nbits);
				csum -= data[index++];
			}

			if(index < size + 5)
				valid = false;
			else if(csum != 0)
				LOG(("do_convert_to7: invalid checksum %02x (should be 0)\n", csum));
		}

		for(int i = 0; i < ff_filler ; i++)
			fputc(0xff, f);
		if(ff_filler > 0) {
			LOG(("do_convert_to7: 0xff filler size=%d\n", ff_filler));
		}
		if(valid) {
			LOG(("do_convert_to7: block type=$%02x size=%d\n", data[2], size));
			VLOG(("do_convert_to7: data="));
		} else
			VLOG(("do_convert_to7: special data="));
		for(int i = 0; i < index ; i++) {
			fputc(data[i], f);
			VLOG(("%02x ", data[i]));
		}
		VLOG(("\n"));

		if(valid && data[2] == 0) {
			char name[] = "01234567.ABC";
                        uint8_t t = data[15];
                        uint8_t u = data[16];
                        memcpy(name, data + 4, 8);
                        memcpy(name + 9, data + 12, 3);
                        for (index = 0; name[index]; index++)
                                if(name[index] < ' ' || name[index] >= 127)
                                        name[index] = '?';
                        LOG(( "do_convert_to7: file \"%s\" type=%s,%s at %imn %is\n",
					name,
					(t==0) ? "bas" : (t==1) ? "dat" : (t==2) ? "bin" : "???",
					(u == 0) ? "a" : (u == 0xff) ? "b" : "?",
					(int)pos / 60, (int)pos % 60));
                }
	}
	return 0;
}

int main(int argc, char *argv[])
{
	cassette_image::ptr cassette;
	const struct cassette_image::Format * const *formats = cassette_default_formats;
	FILE *fin, *fout;
	int ret = 0;

	if(argc != 5) {
		usage();
		return 0;
	}
	fin = fopen(argv[3], "rb");
	if(!fin) {
		fprintf(stderr, "File %s not found.\n",argv[3]);
		return -1;
	}

	auto io = util::stdio_read_write(fin, 0x00);
	if(!io) {
		fprintf(stderr, "Out of memory.\n");
		return -1;
	}

	if(cassette_image::open_choices(std::move(io), "wav", formats, cassette_image::FLAG_READONLY, cassette) !=
			cassette_image::error::SUCCESS) {
		fprintf(stderr, "Invalid format of input file.\n");
		fclose(fin);
		return -1;
	}

	fout = fopen(argv[4], "w");
	if(!fout) {
		fprintf(stderr, "File %s could not be opened.\n",argv[4]);
		fclose(fin);
		return -1;
	}

	if(strcmp(argv[1], "convert"))
		usage();
	else if(!strcmp(argv[2], "to7"))
		do_convert_to7(cassette, fout);
	else if(!strcmp(argv[2], "mo5"))
		ret = do_convert_mo5(cassette, fout);
	else
		usage();
	cassette.reset();
	fclose(fout);
	return ret;
}
